#!/usr/bin/perl
# 
# Copyright (C) 2015 Roland Laifer 
#                    Karlsruhe Institute of Technology (KIT), 
#                    Steinbuch Centre for Computing (SCC)
#                    76128 Karlsruhe, Germany
#                    Roland.Laifer (at) kit.edu
#
# ------------------------------------------------------------------------
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------
#
# $Log: jobstats2email.pl,v $
# Revision 1.5  2015/10/28 16:22:19  rz54
# Re-added check if count and sum > 0 to reduce created output
#
# Revision 1.4  2015/10/28 15:47:52  rz54
# Added localjobdir and fromaddr, other minor corrections
#
# Revision 1.3  2015/10/27 17:36:59  rz54
# Lots of changes to better conform to good perl coding style
#
# Revision 1.2  2015/10/20 14:35:58  rz54
# Added medium IO size to output
#
# Revision 1.1  2015/10/20 14:12:49  rz54
# Initial revision
#
#
# This was developped from show_high_jobstats.pl
#
# $Id: jobstats2email.pl,v 1.5 2015/10/28 16:22:19 rz54 Exp $

use strict;
use warnings;

use Getopt::Std;
our($opt_c, $opt_d, $opt_h, $opt_l, $opt_v, $opt_m, $opt_n, $opt_P);

# get script name
(my $myname = $0)	=~ s#^.*/##;
# set empty values for parameters of config file
my $pdshopts = "";
my $fsysname = "";
my @rsyncopts = ();
my $localjobdir = "";
my $fromaddr = "";
# associative array which stores jobid and cluster name for which values should be reported
undef my %jobid2clus;
# associative array which stores jobid and user for which values should be reported
undef my %jobid2user;
# associative array which stores jobid and email address to which mails should be sent
undef my %jobid2mail;
# associative array which caches accounts and corresponding email addresses 
undef my %user2mail;
# associative array which stores jobid, IO type and sum of counts
undef my %job_type2count;
# associative array which stores jobid, IO type and sum of sums
undef my %job_type2sum;
# associative array which stores jobid, IO type and minimum IO value
undef my %job_type2min;
# associative array which stores jobid, IO type and maximum IO value
undef my %job_type2max;
# email program for sending email
my $mailprg = "/usr/lib/sendmail -t";


sub usage {
    print "Usage: $myname -c file [-d] [-h] [-m] [-n] [-v]\n";
    print "$myname sends an email for each batch job which requested I/O statistic details.\n";
    print "Gets input from Lustre jobstats on all servers and from files created by batch systems.\n";
    print "For more information read the supplied README file and check Roland Laifer's talk at LAD'15,\n";
    print "see http://www.scc.kit.edu/scc/docs/Lustre/kit_lad15_20150922.pdf\n";
    print " -c file       configuration file which descibes how to get files from batch systems.\n";
    print " -d            debug\n";
    print " -h            this help\n";
    print " -l            additionally store sent emails below the directory specified in the config file\n";
    print " -m            send emails (print on STDOUT without this option)\n";
    print " -n            dry-run: do not delete files on batch servers after rsync\n";
    print " -P            omit batch systems which are not reachable via ping (useful for test systems)\n";
    print " -v            verbose\n\n";
    print "Example: $myname -c /etc/$myname.cfg -m\n";
    exit 0;
}


sub parse_cfg {
    open my $cfg_fh, "<", $opt_c or die "*** Can't open $opt_c ($!), stopped";
    while (my $line = <$cfg_fh>) { 
        if ($line =~ /^\s*PDSHOPTS\s*=\s*"(.*)"\s*$/) {
	    $pdshopts = $1;
        } elsif ($line =~ /^\s*FSYSNAME\s*=\s*"(.*)"\s*$/) {
	    $fsysname = $1;
        } elsif ($line =~ /^\s*RSYNCOPTS\s*=\s*"(.*)"\s*$/) {
	    my $rsyncstr = $1;
	    $rsyncstr =~ s/FSYSNAME/$fsysname/g;
	    push(@rsyncopts, "$rsyncstr");
        } elsif ($line =~ /^\s*LOCALJOBDIR\s*=\s*"(.*)"\s*$/) {
	    $localjobdir = $1;
        } elsif ($line =~ /^\s*FROMADDR\s*=\s*"(.*)"\s*$/) {
	    $fromaddr = $1;
        } elsif ($line =~ /^\s*$/ || $line =~ /^\s*\#/ ) {
            # do nothing, just skip empty lines or comment lines
        } else {
            print STDERR "Error scanning line in file $opt_c:\n$line\nExiting...\n";
            exit 1;
        }
    }
    close $cfg_fh;
    if ($opt_d) {
	print "Found config file options in $opt_c:\n";
	print "PDSHOPTS: $pdshopts\n";
	print "FSYSNAME: $fsysname\n";
	print "RSYNCOPTS:\n" . join("\n", @rsyncopts) . "\n";
    }
    if ( ! $pdshopts || ! $fsysname || ! join("", @rsyncopts) || ! $fromaddr || ($opt_l && ! $localjobdir) ) {
	print STDERR "Config file $opt_c is not complete.\nExiting...\n";
	exit 1;
    }
    if ( $fromaddr !~ /\@/) {
	print STDERR "Config file FROMADDR ($fromaddr) does not contain \@ !\n\nExiting...";
	exit 1;
    }
    unless ( -d $localjobdir ) {
	print STDERR "Config file LOCALJOBDIR ($localjobdir) does not exist !\n\nExiting...";
	exit 1;
    }
}


sub get_batchfiles {
    my $cmd;
    while ( my $ropt = shift(@rsyncopts) ) {
	if ($opt_P) {
	    if ( $ropt =~ m|\@([a-z,0-9,\.,-]+):| ) {
		$cmd = 'LANG=C ping -c 1 ' . $1 . ' | grep " 0% packet loss"';
	    } else {
		print STDERR "Found no hostname in RSYNCOPTS ($ropt). Exiting...\n";
		exit 1;
	    }
	    print "Executing: $cmd\n" if $opt_v;
	    my @lines = `$cmd`;
	    my $cmdout = join("", @lines); 
	    unless ( $cmdout ){
		print "Command $cmd had no output. Skipping transfer of files from this batch system.\n" if $opt_v;
		next;
	    }
	}
	if ($opt_n) {
	    $cmd = "rsync -a -e \"ssh -i /root/.ssh/jobstats\" \"$ropt\" $localjobdir 2>&1";
	} else {
	    $cmd = "rsync -a --remove-source-files -e \"ssh -i /root/.ssh/jobstats\" \"$ropt\" $localjobdir 2>&1";
	}
	print "Executing: $cmd\n" if $opt_v;
	my @lines = `$cmd`;
	my $cmdout = join("", @lines); 
	if ( $cmdout ){
	    if ( $cmdout =~ /failed: No such file or directory/ ) {
		print "It appears that command $cmd had found no files. Output:\n$cmdout\n" if $opt_v;
	    } else {
		print STDERR "Command $cmd had unusual output:\n" . join("", @lines) . "Exiting...\n";
		exit 1;
	    }
	}
    }
}


sub parse_batchfiles {
    # 2>/dev/null omits an error in case no files exist
    my $cmd = "ls ${localjobdir}lustrestat-${fsysname}-* 2>/dev/null";
    print "Executing: $cmd\n" if $opt_v;
    my @files = `$cmd`;
    while ( my $file = shift(@files) ) {
        chop $file;
	if ( $file =~ m|${localjobdir}lustrestat-${fsysname}-([a-z,0-9]+)-(\S+)$| ) {
	    my $jobid = $2;
	    $jobid2clus{$jobid} = $1;
	    print "Found file with jobid ($jobid) and cluster ($1): $file\n" if($opt_d);
	    open my $batch_fh, "<", $file or die "*** Can't open $file ($!), stopped";
	    while (my $line = <$batch_fh>) { 
		if ($line =~ /^\s*USER\s*=\s*(\S+)\s*$/) {
		    $jobid2user{$jobid} = $1;
		    print "Found user in file $file: $jobid2user{$jobid}\n" if($opt_d);
		} elsif ($line =~ /^\s*MAIL\s*=\s*(\S+)\s*$/) {
		    $jobid2mail{$jobid} = $1;
		    print "Found mail address in file $file: $jobid2mail{$jobid}\n" if($opt_d);
		} elsif ($line =~ /^\s*MAIL\s*=\s*$/) {
		    # do nothing, just skip empty email entries
		} elsif ($line =~ /^\s*$/ || $line =~ /^\s*\#/ ) {
		    # do nothing, just skip empty lines or comment lines
		} else {
		    print STDERR "Error scanning line in file $file:\n$line\nExiting...\n";
		    exit 1;
		}
	    }
	    close $batch_fh;
	    unless ( $jobid2user{$jobid} ) {
		print STDERR "Jobstats file $file is not complete.\nExiting...\n";
		exit 1;
	    }
	}
    }
    # option -f omits an error in case no files exist
    $cmd = "rm -f ${localjobdir}lustrestat-${fsysname}-*";
    print "Executing: $cmd\n" if $opt_v;
    system("$cmd");
}


sub parse_jobstats {
    my ($cmd, $jobid);
    unless ($opt_v) {
	$cmd = "pdsh $pdshopts -f1 'lctl get_param *." . $fsysname . "-*.job_stats' 2>/dev/null";
	# Note that fanout 1 (-f1) is needed, otherwise output is mixed up. 
	# The " 2>/dev/null" was added since the command reports an error when no Lustre device is mounted.
    } else {
	$cmd = "pdsh $pdshopts -f1 'lctl get_param *." . $fsysname . "-*.job_stats'";
    }
    print "Executing: $cmd\n" if $opt_v;
    my @lines = `$cmd`;
    while ( my $line = shift(@lines) ) {
        chop $line;
	# remove hostname on pdsh command output
	if ( $line =~ /^\S+:\s(.*)$/ ) {
	    $line = $1;
	}
	if ( $line =~ /^-\s+job_id:\s+(\S+)/ ) {
	    $jobid = $1;
	    print "Found line with jobid ($jobid): $line\n" if($opt_d);
	} elsif ( $line =~ /^\s+(\S+):\s+\{\s+samples:\s+(\d+),\s+unit:\s+bytes,\s+min:\s+(\d+),\s+max:\s+(\d+),\s+sum:\s+(\d+)/ ) {
	    if ( $jobid2clus{$jobid} ) {
		my $type = $1;
		my $count = $2;
		my $min = $3;
		my $max = $4;
		my $sum = $5;
		# This is done to omit lines with 0 count in emails.
                if ( $count > 0 ) {
		    if ( $job_type2count{$jobid}{$type} ) {
			$job_type2count{$jobid}{$type} += $count;
		    } else {
			$job_type2count{$jobid}{$type} = $count;
		    }
		}
		if ( ! $job_type2min{$jobid}{$type} || $min < $job_type2min{$jobid}{$type} ) {
		    $job_type2min{$jobid}{$type} = $min;
		}
		if ( ! $job_type2max{$jobid}{$type} || $max > $job_type2max{$jobid}{$type} ) {
		    $job_type2max{$jobid}{$type} = $max;
		}
                if ( $sum > 0 ) {
		    if ( $job_type2sum{$jobid}{$type} ) {
			$job_type2sum{$jobid}{$type} += $sum;
		    } else {
			$job_type2sum{$jobid}{$type} = $sum;
		    }
		}
		print "Found line with samples, type ($type), count ($count), min ($min), max ($max) and sum ($sum): $line\n" if($opt_d);
	    } else {
		print "Skipping line because ($jobid) is not requested: $line\n" if($opt_d);
	    }
	} elsif ( $line =~ /^\s+(\S+):\s+\{\s+samples:\s+(\d+),/ ) {
	    if ( $jobid2clus{$jobid} ) {
		my $type = $1;
		my $count = $2;
                if ( $count > 0 ) {
		    if ( $job_type2count{$jobid}{$type} ) {
			$job_type2count{$jobid}{$type} += $count;
		    } else {
			$job_type2count{$jobid}{$type} += $count;
		    }
		}
		print "Found line with samples, type ($type) and count ($count): $line\n" if($opt_d);
	    } else {
		print "Skipping line because ($jobid) is not requested: $line\n" if($opt_d);
	    }
	} elsif ( $line =~ /^\s+snapshot_time:\s+\d+/ || $line =~ /job_stats:\s*$/ || $line =~ /job_stats=\s*$/ ) {
	    print "Found line without importance: $line\n" if($opt_d);
	} else {
	    print STDERR "Exiting because of unexpected line: $line\n";
	    exit 1;
	}
    }
}


sub get_maddr {
    foreach my $jobid (sort(keys(%jobid2user))) {
	unless ( $jobid2mail{$jobid} ) {
	    my $user = $jobid2user{$jobid};
	    unless ( $user2mail{$user} ) {
		# The following command sends a request to the LDAP server and is site specific 
		my $cmd = "/root/bin/sl -u $user";
		print "Executing: $cmd\n" if $opt_v;
		my @lines = `$cmd`;
		while ( my $line = shift(@lines) ) {
		    if ($line =~ /^mail:\s*(\S+)\s*$/) {
			$user2mail{$user} = $1;
			print "Found email address of user $user: $user2mail{$user}\n" if($opt_d);
		    }
		}
		if ( ! $user2mail{$user} ) {
		    print STDERR "Found no email address of user $user !\n";
		} elsif ( $user2mail{$user} !~ /\@/) {
		    print STDERR "Email address of user $user does not contain \@ !\n";
		    undef $user2mail{$user};
		} else {
		    $jobid2mail{$jobid} = $user2mail{$user};
		}
	    } else {
		$jobid2mail{$jobid} = $user2mail{$user};
	    }
	}
    }
}


# from Perl FAQ: add commas to numbers (makes huge numbers better readable)
sub commify {
    my $input = shift;
    $input = reverse $input;
    $input =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return reverse $input;
}


sub create_output {
    my ($mail_fh, $local_fh, $medsize);
    foreach my $jobid (sort(keys(%jobid2clus))) {
	my $mailaddr;
	my $user = $jobid2user{$jobid};
	if ($opt_m) {
	    unless ( $jobid2mail{$jobid} ) {
		print STDERR "Skipping Job $jobid because $user does not have email address !\n";
		next;
	    } else {
		$mailaddr = $jobid2mail{$jobid};
		open $mail_fh, "|-", $mailprg or die "open of $mailprg failed: $!\n";
		print "Sending email to $mailaddr.\n" if ($opt_v);
	    }
	} else {
	    if ( $jobid2mail{$jobid} ) {
		$mailaddr = $jobid2mail{$jobid};
	    } else {
		$mailaddr = "NONE";
	    }
	    open $mail_fh, ">-" or die "open of stdout failed: $!\n";
	    print "\nTest output of email to $mailaddr:\n\n";
	}
	print $mail_fh "From: $fromaddr\n";
	print $mail_fh "Subject: Lustre IO statistics of your job $jobid on cluster $jobid2clus{$jobid}\n";
	print $mail_fh "To: $mailaddr\n\n";
	print $mail_fh "Hello,\n\nthis is the Lustre IO statistics as requested by user $user on cluster $jobid2clus{$jobid}\n";
        print $mail_fh "for file system $fsysname (command \"df\" shows mount point).\n";
	if ($opt_l) {
	    open $local_fh, ">", ${localjobdir} . "sentmail-" . ${fsysname} . "-" . ${jobid} or die "*** Can't open ${localjobdir}sentmail-${fsysname}-${jobid} ($!), stopped";
	    print $local_fh "From: $fromaddr\n";
	    print $local_fh "Subject: Lustre IO statistics of your job $jobid on cluster $jobid2clus{$jobid}\n";
	    print $local_fh "To: $mailaddr\n\n";
	    print $local_fh "Hello,\n\nthis is the Lustre IO statistics as requested by user $user on cluster $jobid2clus{$jobid}\n";
	    print $local_fh "for file system $fsysname (command \"df\" shows mount point).\n";
	}
	if ( $job_type2count{$jobid} ) {
	    print $mail_fh "\nJob $jobid has done ...\n";
	    foreach my $type (keys %{ $job_type2count{$jobid} }) {
		if ( $job_type2sum{$jobid}{$type} ) {
		    my $commanum = &commify ( $job_type2sum{$jobid}{$type} );
		    if ( $job_type2count{$jobid}{$type} > 0 ) {
			$medsize = int( $job_type2sum{$jobid}{$type} / $job_type2count{$jobid}{$type} );
		    } else {
			$medsize = "NONE";
		    }
		    print $mail_fh "... $job_type2count{$jobid}{$type} $type operations and sum of $commanum byte $type" . "s (min IO size: $job_type2min{$jobid}{$type}, medium IO size: $medsize, max IO size: $job_type2max{$jobid}{$type}).\n";
		    if ($opt_l) {
			print $local_fh "... $job_type2count{$jobid}{$type} $type operations and sum of $commanum byte $type" . "s (min IO size: $job_type2min{$jobid}{$type}, medium IO size: $medsize, max IO size: $job_type2max{$jobid}{$type}).\n";
		    }
		} else {
		    print $mail_fh "... $job_type2count{$jobid}{$type} $type operations.\n";
		    if ($opt_l) {
			print $local_fh "... $job_type2count{$jobid}{$type} $type operations.\n";
		    }
		}
	    }
	} else {
	    print $mail_fh "\nJob $jobid had no Lustre IO statistics !\n";
	    print $mail_fh "\nEither the Lustre IO statistics has already disappeared\n";
	    print $mail_fh "or the job has done no IO on file system $fsysname.\n";
	    if ($opt_l) {
		print $local_fh "\nJob $jobid had no Lustre IO statistics !\n";
		print $local_fh "\nEither the Lustre IO statistics has already disappeared\n";
		print $local_fh "or the job has done no IO on file system $fsysname.\n";
	    }
	}
	close ($mail_fh);
	if ($opt_l) {
	    close ($local_fh);
	}
    }
}


# get options
getopts("c:dhlvmnP") || usage;
usage if ( $opt_h );
usage unless ($#ARGV == -1);
usage unless ($opt_c);

parse_cfg;

get_batchfiles; 

parse_batchfiles;

parse_jobstats;

get_maddr; 

create_output;

exit 0;
